@@ -156,14 +156,59 @@ def NotEqualsInteger(a: AST, b: AST):
 EqualsBool = Iff
 
 # List Utils
+@dataclass(frozen=True)
+class EmptyList(AST):
+    sample_value: uplc_ast.Constant
+
+    def compile(self) -> uplc_ast.AST:
+        return uplc_ast.BuiltinList([], self.sample_value)
+
+    def mk_nil_suffix(self):
+        if isinstance(self.sample_value, uplc_ast.BuiltinPair):
+            return f"Pair<{EmptyList(self.sample_value.l_value).mk_nil_suffix()}|{EmptyList(self.sample_value.r_value).mk_nil_suffix()}>"
+        if isinstance(self.sample_value, uplc_ast.BuiltinList):
+            return f"List{EmptyList(self.sample_value.sample_value).mk_nil_suffix()}"
+        return self.sample_value.__class__
+
+    def dumps(self) -> str:
+        # Note: this is not a real builtin. Essentially, this is not pluto
+        return f"MkNil{self.mk_nil_suffix()} ()"
 
 
-def EmptyList():
+def EmptyIntegerList():
+    return EmptyList(uplc_ast.BuiltinInteger(0))
+
+
+def EmptyByteStringList():
+    return EmptyList(uplc_ast.BuiltinByteString(b""))
+
+
+def EmptyTextList():
+    return EmptyList(uplc_ast.BuiltinString(""))
+
+
+def EmptyBoolList():
+    return EmptyList(uplc_ast.BuiltinBool(False))
+
+
+def EmptyUnitList():
+    return EmptyList(uplc_ast.BuiltinUnit())
+
+
+def EmptyListList(sample_value: uplc_ast.BuiltinList):
+    return EmptyList(sample_value)
+
+
+def EmptyPairList(sample_value: uplc_ast.BuiltinPair):
+    return EmptyList(sample_value)
+
+
+def EmptyDataList():
     # Create an empty list
     return MkNilData(Unit())
 
 
-def EmptyPairList():
+def EmptyDataPairList():
     # Create an empty list of pair type
     return MkNilPairData(Unit())
 
@@ -172,12 +217,12 @@ def EmptyPairList():
 PrependList = MkCons
 
 
-def SingleList(x: AST):
-    return PrependList(x, EmptyList())
+def SingleDataList(x: AST):
+    return PrependList(x, EmptyDataList())
 
 
-def SinglePairList(x: AST):
-    return PrependList(x, EmptyPairList())
+def SingleDataPairList(x: AST):
+    return PrependList(x, EmptyDataPairList())
 
 
 def FoldList(l: AST, f: AST, a: AST):
@@ -250,7 +295,7 @@ def Range(limit: AST, start: AST = Integer(0), step: AST = Integer(1)):
                                 AddInteger(Var("cur"), Var("step")),
                             ),
                         ),
-                        EmptyList(),
+                        EmptyIntegerList(),
                     ),
                 )
             ),
@@ -261,7 +306,7 @@ def Range(limit: AST, start: AST = Integer(0), step: AST = Integer(1)):
     )
 
 
-def MapList(l: AST, m: AST = Lambda(["x"], Var("x"))):
+def MapList(l: AST, m: AST = Lambda(["x"], Var("x")), empty_list=EmptyDataList()):
     """Apply a map function on each element in a list"""
     return Apply(
         Lambda(
@@ -271,7 +316,7 @@ def MapList(l: AST, m: AST = Lambda(["x"], Var("x"))):
                     ["map", "xs"],
                     Ite(
                         NullList(Var("xs")),
-                        EmptyList(),
+                        empty_list,
                         PrependList(
                             Apply(Var("op"), HeadList(Var("xs"))),
                             Apply(
@@ -305,9 +350,9 @@ def NthField(d: AST, n: AST):
 
 
 def NoneData():
-    return ConstrData(Integer(0), EmptyList())
+    return ConstrData(Integer(0), EmptyDataList())
 
 
 def SomeData(x: AST):
     # Note: x must be of type data!
-    return ConstrData(Integer(1), SingleList(x))
+    return ConstrData(Integer(1), SingleDataList(x))
